/*
 
 DESCRIPTION:
 TLOG is a companion program for the data logger used to display & log
 temperatures acquired over the analog channels on a polled basis.  It supports
 any combination of active & NTC sensors using a simple config file which also
 documents program use.
 
 The program is written in ANSI C and should compile without modification on most
 U/L*nux & mac OS platforms.  Provided the include files exist as specified in
 the statements the command 'cc tlog.c' will create the executable.
 
 TLOG communicates using a virtual serial port enumerated as a character
 device in /dev.
 
 This software is licensed under the MIT license:
 
 Copyright 2019 Brian Cornell.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 SYNTAX:        run at command line without options.
 
 KNOWN BUGS:    1) Inherits the same problems as PVLOG and can hang if an
                unexpected / no respnse is received from logger.  Logger's
                BT module also exhibits random lockups or data loss that
                only appear to occur with extended run-times (+24h).  TLOG
                will spot data corruption.
 
 DEPENDENCIES:  tlog.h - program include file
                logger.h - include file for logger firmware
 
 VERSION:  1.0, see build file for build #
 
 MODIFICATION HISTORY:
 
 19 Feb 19, BWC, first release.
 
 
 */


#define _LCLIENT                        //Required for logger client programs
//#define _BSD                          //Uncomment when compiling for FreeBSD

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>

#ifdef _BSD
#include "logger.h"
#else
#include "../logger_1-1.X/logger.h"
#endif

#include "tlog.h"


//Module variables so signal hander has access:  only functions main & sigHandler should directly access.
FILE *LoggerDev = NULL,                 //Logger device handle
     *LogFile = NULL;                   //Log file handle
char *prgName = STR_PRGNAME;


/*Function main.  Program entry point and execution control.
 *
 *
 * Returns:     return code; reference 'RC_' constants in tlog.h.
 *
 * Parameters:  argc=number of command line arguments
 *              argv=array of pointers to argument strings
 *
 */
int main(int argc, char *argv[])
{
    temp  tmp[SIZ_ARY_TMP];             //Object array of temperature data
    logag LogAng;                       //Analog data record
    time_t LogEpoch_t,                  //Time that logger was initialized
           SysTime_t,                   //Used to fetch system time
           calctm;                      //Calculate time from logger epoch
    struct tm *formatm;                 //Formatting of time from logger epoch
    int ReturnCode,                     //Program exit code
        counter,                        //Loop counter
        cflags,                         //Bit map of command line flags
        argc_c;                         //Argument counter
    unsigned int RecCount;              //# records read

    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Argument processing:  If no arguments display program info & exit.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_NOARG;
        if (argc < SIZ_ARGC_MIN)
        {
            fprintf(stdout, FMT_PRG_EXIT);
            fprintf(stdout, FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
            fprintf(stdout, FMT_PRG_SYNTX, prgName);
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)\
                printf(FMT_PRG_CMDL, Flags[counter].str, Flags[counter].desc);
            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Parse command line flags.  The outer loop walks thru the list of
         * arguments and the inner loop scans the array of Flags against each
         * argument.  When a flag is found its respective bit in cflags is set
         * and the value pointer in Flags is set to that parameter.
         *
         * Invalid flags are ignored.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (argc_c = 1, cflags = 0; (argc_c < argc); ++argc_c)
        {
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)
                if (!(strncasecmp(Flags[counter].str, argv[argc_c], \
                                  strlen(Flags[counter].str))))
                {
                    Flags[counter].value = argv[argc_c];
                    break;
                }
            if (counter < SIZ_A_FLG) cflags |= (FLG_BIT_SET << counter);
        }
        
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Load config file.  If the 'different' flag is set skip past the
         * flag portion of the value; otherwise set value to the program name.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DIFF_CONF) Flags[FLG_A_DIFF_CONF].value +=\
            strlen(Flags[FLG_A_DIFF_CONF].str);
        else Flags[FLG_A_DIFF_CONF].value = argv[POS_ARG_PRG];
        if ((ReturnCode = loadCfg(Flags[FLG_A_DIFF_CONF].value, \
                                  (cflags & FLG_DIFF_CONF), tmp, \
                                  SIZ_ARY_TMP))) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If dump flag set, dump config & exit
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DUMP_CONF)
        {
            for (counter = 0; (counter < SIZ_ARY_TMP); ++counter)
                fprintf(stdout, FMT_CFG_DUMP, counter, tmp[counter].lbl,\
                        tmp[counter].ntc, tmp[counter].mvc, tmp[counter].zof,\
                        tmp[counter].ce1, tmp[counter].ce2, tmp[counter].ce3,\
                        tmp[counter].ce4);
            ReturnCode = RC_SUCCESS;
            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If log data flag set, open file
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_LOG_FILE)
        {
            ReturnCode = RC_LOGFAIL;
            Flags[FLG_A_LOG_FILE].value += strlen(Flags[FLG_A_LOG_FILE].str);
            if ((LogFile = fopen(Flags[FLG_A_LOG_FILE].value, \
                                 FMODE_LOG_FIL)) == NULL) break;
            if (!(fprintf(LogFile, FMT_HDR_LOG))) break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                        Finish initialization
         * Open log device or file; exit on fail.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_DEVOFAIL;
        if ((LoggerDev = fopen(argv[POS_ARG_DEV], FMODE_LOG_DEV)) == NULL) break;

        //Install signal handlers
        ReturnCode = RC_SIGERROR;
        if (signal(SIGINT, sigHandler) == SIG_ERR) break;
        if (signal(SIGTERM, sigHandler) == SIG_ERR) break;
        if (signal(SIGHUP, sigHandler) == SIG_ERR) break;
        if (signal(SIGQUIT, sigHandler) == SIG_ERR) break;

        ReturnCode = RC_DEVDFAIL;
        if (startLog(LoggerDev) == FALSE) break;

        //Read header record & validate; exit on fail
        ReturnCode = RC_LOGHFAIL;
        if ((LogEpoch_t = readHeader(LoggerDev)) == (time_t) -1) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Setup display
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ansiClrScr();
        ansiCursorPos(POS_HDR_R, POS_HDR_C);
        fprintf(stdout, FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
        ansiCursorPos(POS_HDR_LBL_R, POS_HDR_LBL_C);
        fprintf(stdout, FMT_HDR_LABEL);
        ansiCursorPos(POS_AMB_LBL_R, POS_AMB_LBL_C);
        fprintf(stdout, FMT_LBL_AMB, tmp[MAP_TMP_AMB].lbl);
        ansiCursorPos(POS_CH2_LBL_R, POS_CH2_LBL_C);
        fprintf(stdout, FMT_LBL_CH2, tmp[MAP_TMP_CH2].lbl);
        ansiCursorPos(POS_CH4_LBL_R, POS_CH4_LBL_C);
        fprintf(stdout, FMT_LBL_CH4, tmp[MAP_TMP_CH4].lbl);
        ansiCursorPos(POS_CH5_LBL_R, POS_CH5_LBL_C);
        fprintf(stdout, FMT_LBL_CH5, tmp[MAP_TMP_CH5].lbl);
        ansiCursorPos(POS_SEQ_LBL_R, POS_SEQ_LBL_C);
        fprintf(stdout, FMT_LBL_SEQ);
        ansiCursorPos(POS_CMD_R, POS_CMD_C);
        fprintf(stdout, FMT_MSG_EXIT);
        if (LogFile) fprintf(stdout, FMT_MSG_LOG, Flags[FLG_A_LOG_FILE].value);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Display temps until signal received or error in data stream.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (RecCount = 0, ReturnCode = RC_LOGAFAIL;\
             (fread((void *) &LogAng, sizeof(logag), 1, LoggerDev)); ++RecCount,\
             ReturnCode = RC_LOGAFAIL)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Data validation checks:
             *  -Record type
             *  -Sequence number
             *  -Analog channel
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            ReturnCode = RC_RECTYPE;
            if (LogAng.rectype != RECTANL)                  //Record type
            {
                //Byte-byte data dump for diagnostics
                int c;
                unsigned char *p;
                fprintf(stderr, FMT_ERR_RECT);
                for (c = 0, p = (unsigned char *) &LogAng; (c < sizeof(logag));\
                     ++c, ++p) fprintf(stderr, FMT_ERR_RECTV, (unsigned int) *p);
                break;
            }
            if (LogAng.seqnum != RecCount)                  //Sequence number
            {
                ReturnCode = RC_SEQNUMBR;
                fprintf(stderr, FMT_ERR_SEQ, RecCount, \
                        (unsigned int) LogAng.seqnum);
                break;
            }
            //Analog channel
            if ((LogAng.pin < MAP_TMP_AMB) || (LogAng.pin > MAP_TMP_CH5))
            {
                ReturnCode = RC_INVCHANL;
                break;
            }

            //Validation complete - analyze / process record.
            ansiCursorSav();                        //Save before update
            ansiCursorPos(tmp[LogAng.pin].row, tmp[LogAng.pin].col);
            ansiEraseLine();
            getTemp(LogAng.adcval, LogAng.adcv, &tmp[LogAng.pin]);
            fprintf(stdout, FMT_TMP_DSP, tmp[LogAng.pin].now, \
                    tmp[LogAng.pin].min, tmp[LogAng.pin].max, \
                    tmp[LogAng.pin].dif);

            //Print sequence number (helps to know if hung) & supply voltage (Vdd)
            ansiCursorPos(POS_SEQ_VAL_R, POS_SEQ_VAL_C);
            ansiEraseLine();
            fprintf(stdout, FMT_SEQ_DSP, RecCount, FMT_LBL_VDD, \
                    (float) ((float) ADC_MAX_VAL / (float) LogAng.adcv));
            ansiCursorRst();                        //Restore cursor position

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Log data to file if option set
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (LogFile)
            {
                calctm = LogEpoch_t + (time_t) TimeSinceEpoch(&LogAng.elapsed);
                formatm = localtime(&calctm);
                if (!(fprintf(LogFile, FMT_OUT_LOG, LogAng.seqnum, \
                              (formatm->tm_mon+1), formatm->tm_mday, \
                              (formatm->tm_year + SYS_EPOCH_YEAR),\
                              formatm->tm_hour, formatm->tm_min, \
                              formatm->tm_sec, LogAng.pin, tmp[LogAng.pin].lbl,\
                              tmp[LogAng.pin].now, tmp[LogAng.pin].min, \
                              tmp[LogAng.pin].max, tmp[LogAng.pin].dif)))
                {
                    ReturnCode = RC_LOGFAIL;
                    break;
                }
            }
        }
        printf("\n");
        break;
    }

    //Clean-up & exit
    if (LoggerDev) stopLog(LoggerDev);
    if (LoggerDev) fclose(LoggerDev);
    if (LogFile) fclose(LogFile);
    if (ReturnCode > RC_NOARG) fprintf(stderr, FMT_ERR_STDERR, prgName,\
                                       retCode[ReturnCode], errno);
    fprintf(stdout, FMT_PRG_EXIT);

    return(ReturnCode);
} //End function main


/*Function getTemp.  Calculates the temperature represented by the adc value.
 *
 *
 * Returns:     n/a.  Object pointed to by channel contains updated temps
 *
 * Parameters:  adc=value read by adc
 *              onev=adc value for one volt
 *              channel=object describing sensor type & storage for temps
 *
 */
void getTemp(int adc, int onev, temp *channel)
{
    float volts;

    volts = ((float) adc / (float) onev);
    if (channel->ntc)
        channel->now = (channel->ce1 * pow(volts, 3) + channel->ce2 * \
                        pow(volts, 2) - channel->ce3 * volts + channel->ce4);
    else
        channel->now = ((volts / ((float) channel->mvc * (float) 0.001)) - \
                        (float) channel->zof);

    if (channel->now < channel->min)
        channel->min = channel->now;
    if (channel->now > channel->max)
        channel->max = channel->now;
    channel->dif = (channel->max - channel->min);

    return;
} //End function getTemp


/*Function getVolts.  Calculates the voltage represented by the adc value and scales.
 *
 *
 * Returns:     float value representing voltage.
 *
 * Parameters:  adc=value read by adc
 *              onev=adc value for one volt
 *              scale=integer representing the scale of the adc value
 *
 */
float getVolts(int adc, int onev, int scale)
{
    return(((float) ( (float) adc / (float) onev) * (float) scale));
} //End function getVolts


/*Function getAmps.  Calculates the amps represented by the adc value.
 *
 *
 * Returns:     float value representing amperes.
 *
 * Parameters:  adc=value read by adc
 *              onev=adc value for one volt
 *              offset=adc value representing 0A
 *              mVamp=float value representing sensor's mV per amp scale
 *
 */
float getAmps(int adc, int onev, int offset, float mVamp)
{
    return((getVolts(adc, onev, 1) - getVolts(offset, onev, 1)) / mVamp);
} //End function getAmps


/*Function TimeSinceEpoch.  Calculates the number of seconds since logger's epoch.
 *
 *
 * Returns:     integer representing the number of seconds since epoch 
 *              representing by object tse.
 *
 * Parameters:  tse=pointer to object of type logtm representing the epoch.
 *
 */
unsigned int TimeSinceEpoch(logtm *tse)
{
    return((tse->Min * 60) + tse->Sec);
} //End function TimeSinceEpoch


/*Function startLog.  Attempts to establish a known working state with logger and
 * initiate logging.  This is problematic mainly because there is no facility to
 * determine if data is waiting to be read from the stream.  Since the stream 
 * read functions are blocking they don't return until a character is ready.
 *
 * This function worsk around this by sending a series of commands that attempt 
 * to ensure that the logger is in command mode and binary output is on.  The ID
 * command is issued so that the last expected response is the processor (PIC MPU)
 * ID.  The output stream is read seaching for this to avoid reading past the end
 * of available data (to avoid a hung state) and confirm that logger is waiting 
 * for a command.  Once confirmed the start logging command is issued.
 *
 *
 * Returns:     TRUE if start logging was successful
 *              FALSE if logging could not be started; logger state unknown
 *
 * Parameters:  out is a pointer to an object of type FILE representing the device
 *              stream to communicate with logger
 *
 */
int startLog(FILE *out)
{
    time_t systime;
    struct tm *ltime;
    unsigned char idstr[SIZ_BUF_IDS];
    char cmdrsp[SIZ_RSP_CMD];
    int cmdpos, idpos, retval;

    for (retval = FALSE; (TRUE);)                           //Waterfall loop
    {
        if (snprintf((char *) idstr, (size_t) SIZ_BUF_IDS, FMT_BUF_IDS,\
                     (unsigned int) PROC_ID) < 0) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * First ensure logging is stopped and set binary mode.  Issue ID
         * command.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (fprintf(out, FMT_LOG_HDSK) < 0) break;
        for (cmdpos=SIZ_RSP_CMD, idpos = 0; (cmdpos); --cmdpos, ++idpos)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Look for processor ID; this marks end of logger output in command 
             * mode and puts it in known good state to issue further commands.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((char) fgetc(out) != idstr[idpos]) idpos = 0;
            if (idpos == (SIZ_PROC_ID - 1)) break;
        }
        if (!(cmdpos)) break;       //Didn't find ID - unknown response / state!

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Fetch system time and update logger epoch.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        time(&systime);
        ltime = localtime(&systime);
        if (fprintf(out, FMT_LOG_EPOCH,\
                    (ltime->tm_year - ((ltime->tm_year / 100) * 100)),\
                    (ltime->tm_mon + 1), ltime->tm_mday, ltime->tm_hour,\
                    ltime->tm_min, ltime->tm_sec) < 0) break;
        fgets(cmdrsp, SIZ_RSP_CMD, out);            //Discard response

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Configure low voltage shutdown
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (fprintf(out, FMT_LOG_LVS) < 0) break;
        fgets(cmdrsp, SIZ_RSP_CMD, out);            //Discard response

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Configure channels
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (fprintf(out, FMT_LOG_CONF) < 0) break;
        fgets(cmdrsp, SIZ_RSP_CMD, out);            //Discard response

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Start logging
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (fprintf(out, FMT_LOG_START) < 0) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Consume command response up to start of header record.  Push first
         * byte back onto stream.  If not found in size of header record bytes
         * then problem.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (cmdpos = SIZ_RSP_CMD; (cmdpos); --cmdpos)
            if (fgetc(out) == (int) RECTHDR) break;
        if (!(cmdpos)) break;
        ungetc((int) RECTHDR, out);

        retval = TRUE;
        break;
    }

    return(retval);
} //End function startLog


/*Function stopLog.  Attempts to stop logging and confirm logger is in command
 * mode.  This is problematic mainly because there is no facility to determine
 * if data is waiting to be read from the stream.  Since the stream read functions
 * are blocking they don't return until a character is ready.
 *
 * After the stop logging command is sent, the ID command is issued so that the
 * last expected response is the processor (PIC MPU) ID.  The output stream is
 * read seaching for this to avoid reading past the end of available data (to
 * avoid a hung state) and confirm that logger is waiting for a command.
 *
 * Returns:     TRUE if successful
 *              FALSE if logging could not be stopped; logger state unknown
 *
 * Parameters:  out=pointer to an object of type FILE representing the device
 *              stream to communicate with logger
 *
 */
int stopLog(FILE *out)
{
    unsigned char idstr[SIZ_BUF_IDS];
    int cmdpos, idpos, retval;
    
    //Return logger to command mode
    for (retval = FALSE; (TRUE);)                           //Waterfall loop
    {
        if (snprintf((char *) idstr, (size_t) SIZ_BUF_IDS, FMT_BUF_IDS, \
                     (unsigned int) PROC_ID) < 0) break;
        
        if (fprintf(out, FMT_LOG_STOP) < 0) break;

        //Wait for logging to stop before issuing ID command
        sleep(WAIT_LOG_STOP);
        if (fprintf(out, FMT_LOG_GID) < 0) break;

        //Now parse responses
        for (cmdpos=SIZ_RSP_CMD, idpos = 0; (cmdpos); --cmdpos, ++idpos)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Look for processor ID; this marks end of logger output & confirms 
             * it's in command mode.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((char) fgetc(out) != idstr[idpos]) idpos = 0;
            if (idpos == (SIZ_PROC_ID - 1)) break;
        }
        if (!(cmdpos)) break;                               //This means ID string wasn't found - unknown response / state!

        retval = TRUE;
        break;
    }
    
    return(retval);
} //End function stopLog


/*Function sigHandler.  Signal handler for user interrupt & TERM signals.  Places
 * logger in command mode & closes device.
 *
 * Returns:     n/a
 *
 * Parameters:  signal=signal #; refer to signal.h.
 *
 */
void sigHandler(int signal)
{
    static int recurse = 1;

    if (recurse)
    {
        recurse = 0;
        stopLog(LoggerDev);
        fclose(LoggerDev);

        if (LogFile) fclose(LogFile);

        fprintf(stderr, FMT_TRM_SIGNAL, prgName, signal);
        exit(RC_SIGEXIT);
    }

    return;
} //End function sigHandler


/*Function readHeader.  Reads logger header from stream, validates, and converts
 * logger's epoch to calendar time.
 *
 * Returns:     On success a time_t value representing logger epoch in calendar 
 *              time.  Otherwise -1 which could indicate a file operation
 *              error, bad header data, or failure to convert the epoch to a
 *              time_t value.  errno should be checked before & after the
 *              function call for detailed error handling.
 *
 * Parameters:  lf=pointer to object of type FILE representing the stream
 *              that the header will be read from.
 *
 */
time_t readHeader(FILE *lf)
{
    time_t SysTime_t;                               //Used to fetch system time
    struct tm LogEpoch;                             //Time that logger was initialized
    loghd LogHdr;                                   //Header record
    time_t retval = -1;

    while (TRUE)
    {
        //Read header record & validate
        if (fread((void *) &LogHdr, sizeof(loghd), SIZ_REC_HDR, lf) != SIZ_REC_HDR) break;
        if (LogHdr.rectype != RECTHDR) break;
        if (LogHdr.deviceid.word != PROC_ID) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Fetch logger epoch, store, and get tm value as local time.  First get
         * current system time so that DST flag can be properly set.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        time(&SysTime_t);
        LogEpoch.tm_isdst = localtime(&SysTime_t)->tm_isdst;
        LogEpoch.tm_sec = (int) LogHdr.epoch.second;
        LogEpoch.tm_min = (int) LogHdr.epoch.minute;
        LogEpoch.tm_hour = (int) LogHdr.epoch.hour;
        LogEpoch.tm_year = (int) ((LogHdr.epoch.year + LOG_EPOCH_YEAR) - SYS_EPOCH_YEAR);
        LogEpoch.tm_mon = (int) (LogHdr.epoch.month - 1);
        LogEpoch.tm_mday = (int) LogHdr.epoch.day;

        retval = mktime(&LogEpoch);
        break;
    }
    
    return(retval);
} //End function readHeader



/*Function ansiClrScr.  Clears terminal screen.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiClrScr(void)
{
    printf("%s%s", ANSIESC, ANSICLRSCR);
    return;
} //End function ansiClrScr


/*Function ansiEraseLine.  Erases line fromo cursor position to end of line.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiEraseLine(void)
{
    printf("%s%c", ANSIESC, ANSIELINE);
    return;
} //End function ansiEraseLine


/*Function ansiCursorSav.  Tells terminal to save terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorSav(void)
{
    printf(ANSISAVE);
    return;
} //End function ansiCursorSav


/*Function ansiCursorRst.  Tells terminal to restore terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorRst(void)
{
    printf(ANSIRESTORE);
    return;
} //End function ansiCursorRst


/*Function ansiCursorPos.  Sets terminal cursor position.
 *
 * Returns:     n/a
 *
 * Parameters:  row=terminal row
 *              col=terminal column
 *
 */
void ansiCursorPos(unsigned char row, unsigned char col)
{
    char srow[ACP_SZ_STR], scol[ACP_SZ_STR];
    
    if ((row < ACP_MAX_ROW) && (col < ACP_MAX_COL))
    {
        snprintf(srow, ACP_SZ_STR, "%hhu", row);
        snprintf(scol, ACP_SZ_STR, "%hhu", col);
        printf("%s%s;%s%c", ANSIESC, srow, scol, ANSICPOS);
    }
    return;
} //End function ansiCursorPos


/*Function loadCfg.  Parses config file and initializes array of temp objects.
 *
 * File format:     ch:lbl;ntc;mvc;zof;ce1;ce2;ce3;ce4
 *                      ch=channel number
 *                          1=built-in ambient sensor
 *                          2=channel 2
 *                          4=channel 4
 *                          5=channel 5
 *                      lbl=display name for channel (max 8 char)
 *                      ntc=1 if NTC or 0 if active sensor
 *                      mvc=active sensor mV/C (integer)
 *                      zof=active sensor offset for 0C, degrees C (integer)
 *                      ce1=ntc coefficient #1, float
 *                      ce2=ntc coefficient #2, float
 *                      ce3=ntc coefficient #3, float
 *                      ce4=ntc coefficient #4, float
 *
 *                  The colon separating ch & lbl is mandatory.  Subsequent
 *                  fields are separated with a semicolon.
 *
 * Returns:         RC_SUCCESS if loaded, otherwise RC_BADCFG
 *
 * Parameters:      cfg=config file to load
 *                  cfg_d=FALSE to load default config; otherwise load cfg
 *                      without modification
 *                  temp_ar=pointer to array of temp objects
 *                  temp_sz=size of temp_ar array
 *
*/
int loadCfg(char *cfg, int cfg_d, temp *temp_ar, int temp_sz)
{
    FILE *cfgfile;                      //Config file handle
    int ch,                             //Channel #
        line,                           //Config file line #
        rc;                             //Return code
    char cfgrec[SIZ_BUF_REC];           //Buffer for config file path
    
    for (cfgfile = NULL, rc = RC_BADCFG; (TRUE);)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Open config file & initialize memory objects.  Use cfgrec to format
         * file path / name.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (strlcpy(cfgrec, cfg, SIZ_BUF_REC) > SIZ_BUF_REC) break;
        if (!(cfg_d))
            if (strlcat(cfgrec, STR_EXT_CFG, SIZ_BUF_REC) > SIZ_BUF_REC) break;
        if (!(cfgfile = fopen(cfgrec, FMODE_CONF))) break;

        //Clear array
        memset((void *) temp_ar, CHR_NULL, (temp_sz * sizeof(temp)));

        //Display row & columns are hard-coded
        (temp_ar + MAP_TMP_AMB)->row = POS_AMB_VAL_R;
        (temp_ar + MAP_TMP_AMB)->col = POS_AMB_VAL_C;
        (temp_ar + MAP_TMP_CH2)->row = POS_CH2_VAL_R;
        (temp_ar + MAP_TMP_CH2)->col = POS_CH2_VAL_C;
        (temp_ar + MAP_TMP_CH4)->row = POS_CH4_VAL_R;
        (temp_ar + MAP_TMP_CH4)->col = POS_CH4_VAL_C;
        (temp_ar + MAP_TMP_CH5)->row = POS_CH5_VAL_R;
        (temp_ar + MAP_TMP_CH5)->col = POS_CH5_VAL_C;

        //Set max & min values to opposite limits for proper update
        (temp_ar + MAP_TMP_AMB)->min = LIM_TMP_MAX;
        (temp_ar + MAP_TMP_AMB)->max = LIM_TMP_MIN;
        (temp_ar + MAP_TMP_CH2)->min = LIM_TMP_MAX;
        (temp_ar + MAP_TMP_CH2)->max = LIM_TMP_MIN;
        (temp_ar + MAP_TMP_CH4)->min = LIM_TMP_MAX;
        (temp_ar + MAP_TMP_CH4)->max = LIM_TMP_MIN;
        (temp_ar + MAP_TMP_CH5)->min = LIM_TMP_MAX;
        (temp_ar + MAP_TMP_CH5)->max = LIM_TMP_MIN;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Process config file records.  Ignore those starting with invalid
         * channel #.  A channel config can be set multiple times with the last
         * one read being applied.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (line = 0; (fgets(cfgrec, SIZ_BUF_REC, cfgfile)); ++line)
        {
            //Get rid of trailing CR/LF
            strnl((unsigned char *) cfgrec, SIZ_BUF_REC);
            ch = strtol(cfgrec, NULL, BASE_TEN);

            //Skip if first field isn't valid channel #
            if ((ch < MAP_TMP_MIN) || (ch > MAP_TMP_MAX)) continue;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Assign fields
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            (temp_ar + ch)->ch = ch;
            strlcpy((temp_ar + ch)->lbl, getCfgVar(cfgrec, POS_CFG_LBL), \
                    SIZ_BUF_LBL);
            (temp_ar + ch)->ntc = strtol(getCfgVar(cfgrec, POS_CFG_NTC), NULL, \
                                         BASE_TEN);
            (temp_ar + ch)->mvc = strtol(getCfgVar(cfgrec, POS_CFG_MVC), NULL, \
                                         BASE_TEN);
            (temp_ar + ch)->zof = strtol(getCfgVar(cfgrec, POS_CFG_ZOF), NULL, \
                                         BASE_TEN);
            (temp_ar + ch)->ce1 = strtof(getCfgVar(cfgrec, POS_CFG_CE1), NULL);
            (temp_ar + ch)->ce2 = strtof(getCfgVar(cfgrec, POS_CFG_CE2), NULL);
            (temp_ar + ch)->ce3 = strtof(getCfgVar(cfgrec, POS_CFG_CE3), NULL);
            (temp_ar + ch)->ce4 = strtof(getCfgVar(cfgrec, POS_CFG_CE4), NULL);
        }
        
        if (cfgfile) fclose(cfgfile);
        break;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Very simple check to validate all array objects loaded but, no guarentee
     * that all elements are valid.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if ((temp_ar + MAP_TMP_AMB)->ch + (temp_ar + MAP_TMP_CH2)->ch + \
        (temp_ar + MAP_TMP_CH4)->ch + (temp_ar + MAP_TMP_CH5)->ch == SUM_TMP_CHL)
        rc = RC_SUCCESS;
    
    return(rc);
} //End function loadCfg


/*Function getCfgVar.  Parses a config record for the given variable.  Not
 * thread safe!
 *
 * Returns:     pointer to NULL terminated string; will be 0 length if variable
 *              not found, 0 length, or corrupt record
 *
 * Parameters:  rec=pointer to NULL terminated config record
 *              pos=offset-1 number representing variable position; if set to
 *              0 then the record identifier & first variable will be returned
 *
 */
const char *getCfgVar(char *rec, int pos)
{
    static char var[SIZ_BUF_VAR];
    int rec_off,
    var_off,
    pos_cur,
    fndc;
    
    memset((void *) var, CHR_NULL, SIZ_BUF_VAR);        //Clear
    for (rec_off = 0, var_off = 0, pos_cur = 0, fndc = 0; \
         (*(rec + rec_off)); ++rec_off)
    {
        if (pos_cur == pos)
        {
            if ((*(rec + rec_off) != CHR_NULL) && \
                (*(rec + rec_off) != CHR_SEMICOLON) && \
                (var_off < (SIZ_BUF_VAR - 1)))
            {
                var[var_off] = *(rec + rec_off);
                ++var_off;
                continue;
            }
            else break;
        }
        
        //First look for colon to find start of first variable
        if ((*(rec + rec_off) == CHR_COLON) && (!(fndc)))
        {
            fndc = TRUE;
            ++pos_cur;                                  //Found 1st variable
            continue;
        }
        
        if (*(rec + rec_off) == CHR_SEMICOLON) ++pos_cur;    //Found next variable
    }
    
    return((const char *) var);
} //End function getCfgVar


/*Function strnl.  Removes the trailing New Line (NL) or Carriage Return (CR)
 *  from a string up to maxsz - 1 (whichever is encountered first) and replaces
 *  with a NULL byte.  In the event that buf was not NULL terminated and NL or
 *  CR aren't found buf will be left unterminated; therefore the calling
 *  function should check the return code.
 *
 *
 * Returns:     length of NULL terminated buffer (minus NULL byte) or zero if CR
 *              or LF isn't found.  If a NULL is encountered first the function
 *              will leave the buffer unmodified and return the length of NULL
 *              terminated buffer (minus NULL byte).
 *
 * Parameters:  buf=pointer to buffer
 *              maxsz=size of buffer
 *
 */
int strnl(unsigned char *buf, int maxsz)
{
    int pos;
    
    for (pos = 0; (pos < maxsz); ++pos)
    {
        if ((*(buf + pos) == CHR_CR) || (*(buf + pos) == CHR_LF) || \
            (*(buf + pos) == CHR_NULL)) break;
    }
    if (pos < maxsz)
    {
        *(buf + pos) = CHR_NULL;
        return(pos);
    }
    return(0);
} //End function strnl



//End file.
